#include "deformation.hpp"

using namespace cgp;

#ifdef SOLUTION
void build_matrix(linear_system_structure& linear_system, constraint_structure const& constraints, mesh const& shape, numarray<vec3> const& initial_position, numarray<numarray<int> > const& one_ring)
{
	size_t const N = shape.position.size();
	size_t const N_constraints = constraints.fixed.size() + constraints.target.size();
	size_t const N_equations = N + N_constraints;

	// initialize right-hand-size vector
	linear_system.rhs_x.resize(N_equations);
	linear_system.rhs_y.resize(N_equations);
	linear_system.rhs_z.resize(N_equations);

	linear_system.guess_x.resize(N);
	linear_system.guess_y.resize(N);
	linear_system.guess_z.resize(N);

	for(size_t k=0; k<shape.position.size(); ++k){
        vec3 const& p = initial_position[k];
        linear_system.guess_x[k] = p.x;
        linear_system.guess_y[k] = p.y;
        linear_system.guess_z[k] = p.z;
    }

	linear_system.M.resize(N_equations, N);
    linear_system.M.reserve(Eigen::VectorXi::Constant(N_equations*N,6));

	// Differential coordinates
    for(int k=0; k<N; ++k)
    {
        const auto& neighbors = one_ring[k];
        const int Nn = neighbors.size();

        vec3 d = shape.position[k];
        for(auto n : neighbors){
            d -= 1.0f/Nn * shape.position[n];
            linear_system.M.coeffRef(k,n) = -1.0f/Nn;
        }

        linear_system.M.coeffRef(k,k) = 1;
    }

	// Constraints
    int k_constraint = 0;
    for(const auto& c : constraints.fixed) {
        int const idx = c.first;
        linear_system.M.coeffRef(N+k_constraint, idx) = constraints.weight_fixed;
        ++k_constraint;
    }
    for(const auto& c : constraints.target) {
        int const idx = c.first;
        linear_system.M.coeffRef(N+k_constraint, idx) = constraints.weight_target;
        ++k_constraint;
    }

	linear_system.M.makeCompressed();

	// Factorization
    linear_system.solver.compute(linear_system.M);
    linear_system.solver.setTolerance(1e-6f);

}
#else
void build_matrix(linear_system_structure& linear_system, constraint_structure const& constraints, mesh const& shape, numarray<vec3> const& initial_position, numarray<numarray<int> > const& one_ring)
{
    // TO DO: Build and fill the matrix M and the rhs (M q_x/y/z = rhs_x/y/z)
    //
    //  linear_system.rhs_x.resize(...)
    //  for all vertices
    //      linear_system.rhs_x[k] = ....
    //  
    //  Fill the matrix:
    //  linear_system.M.resize(...), linear_system.M.reserve(...)
    //  for all non-zero elements:
    //      linear_system.M.coeffRef(i,j) = ...
    // 
}
#endif

#ifdef SOLUTION
void update_deformation(linear_system_structure& linear_system, constraint_structure const& constraints, mesh& shape, mesh_drawable& visual, numarray<vec3> const& initial_position, numarray<numarray<int> > const& one_ring, numarray<mat3>& rotations_arap, bool arap)
{
	size_t const N = shape.position.size();
    size_t N_iteration = 4;

    if(arap==false) N_iteration=1;

	// Need to iterate on rotation in the case of ARAP
	for(int k_iteration=0; k_iteration<N_iteration; ++k_iteration)
	{ 
        if(arap){
		    for(int k=0; k<N; ++k)
            {
			    const auto& neighbors = one_ring[k];
			    // Covariance matrix
                Eigen::Matrix3f cov; cov << 0,0,0, 0,0,0, 0,0,0;
                vec3 const& pk  = shape.position[k];
                vec3 const& pk0 = initial_position[k];
                for(auto n : neighbors){
                    vec3 const& pn  = shape.position[n];
                    vec3 const& pn0 = initial_position[n];

                    float const nn = norm(pn-pk);
                    vec3 const e = (pk-pn)/nn;
                    vec3 const e0 = (pk0-pn0)/nn;


                    Eigen::Matrix3f m;
                    m << e.x*e0.x, e.x*e0.y, e.x*e0.z,
                            e.y*e0.x, e.y*e0.y, e.y*e0.z,
                            e.z*e0.x, e.z*e0.y, e.z*e0.z;

                    cov += m;

                }

			    // Polar decomposition
                Eigen::JacobiSVD<Eigen::Matrix3f> svd(cov, Eigen::ComputeThinU | Eigen::ComputeThinV);
                Eigen::Matrix3f const R = svd.matrixU()* (svd.matrixV().transpose());

                rotations_arap[k] = {R(0,0),R(0,1),R(0,2), R(1,0),R(1,1),R(1,2), R(2,0),R(2,1),R(2,2)};

		    }
        }


		// Differential coordinates
        for(int k=0; k<N; ++k)
        {
            auto const& neighbors = one_ring[k];
            size_t const Nn = neighbors.size();

            vec3 d = initial_position[k];
            for(auto n : neighbors){
                d -= 1.0f/Nn * initial_position[n];
            }
            if(arap)
                d = rotations_arap[k] * d;

            linear_system.rhs_x[k] = d.x;
            linear_system.rhs_y[k] = d.y;
            linear_system.rhs_z[k] = d.z;
        }

		// Add constraints
        int k_constraint = 0;
        for(const auto& c : constraints.fixed)
        {
            vec3 const& p = c.second;
            float const w = constraints.weight_fixed;
            linear_system.rhs_x[N+k_constraint] = w * p.x;
            linear_system.rhs_y[N+k_constraint] = w * p.y;
            linear_system.rhs_z[N+k_constraint] = w * p.z;
            ++k_constraint;
        }
        for(const auto& c : constraints.target)
        {
            vec3 const& p = c.second;
            float const w = constraints.weight_target;
            linear_system.rhs_x[N+k_constraint] = w * p.x;
            linear_system.rhs_y[N+k_constraint] = w * p.y;
            linear_system.rhs_z[N+k_constraint] = w * p.z;
            ++k_constraint;
        }



		// Solving the linear system
        const auto qx = linear_system.solver.solveWithGuess(linear_system.rhs_x, linear_system.guess_x);
        const auto qy = linear_system.solver.solveWithGuess(linear_system.rhs_y, linear_system.guess_y);
        const auto qz = linear_system.solver.solveWithGuess(linear_system.rhs_z, linear_system.guess_z);

        // Update future guess solution
        linear_system.guess_x = qx;
        linear_system.guess_y = qy;
        linear_system.guess_z = qz;

        // Update surface
        for(int k=0; k<N; ++k)
            shape.position[k] = {qx[k], qy[k], qz[k]};

	}

    shape.normal_update();
    visual.vbo_position.update(shape.position);
    visual.vbo_normal.update(shape.normal);

}

#else

void update_deformation(linear_system_structure& linear_system, constraint_structure const& constraints, cgp::mesh& shape, cgp::mesh_drawable& visual, cgp::numarray<cgp::vec3> const& initial_position, cgp::numarray<cgp::numarray<int> > const& one_ring)
{
    // TO DO: Update the RHS with new constraints and Solve the system
    //
    // For all vertex and contraints
    //   linear_system.rhs_x[k] = ... (+ rhs_y, rhs_z)
    // 
    // linear_system.solver.solve(linear_system.rhs_x) ... or linear_system.solver.solveWithGuess(linear_system.rhs_x, linear_system.guess_x), etc.
}
#endif